import json
import logging
from typing import Optional

import botocore
from pscr_common.aws import get_boto3_client

_logger = logging.getLogger(__name__)


class S3Manager:
    s3_client = get_boto3_client("s3")

    @staticmethod
    def get_validated_overrides(
        manual_overrides: str, manual_overrides_backup: str, s3_bucket: str
    ) -> Optional[dict]:
        def _restore_backup() -> Optional[dict]:
            s3_backup_data = S3Manager.get_json_object(
                manual_overrides_backup, s3_bucket
            )
            if not S3Manager._validate_json_object(s3_backup_data):
                _logger.error(
                    f"Neither {manual_overrides} nor {manual_overrides_backup} are valid."
                )
                return None

            _logger.warning(
                f"Restoring override file {manual_overrides} from {manual_overrides_backup}"
            )
            S3Manager.upload_json_object(s3_backup_data, manual_overrides, s3_bucket)
            return s3_backup_data

        # noinspection PyBroadException
        try:
            s3_data = S3Manager.get_json_object(manual_overrides, s3_bucket)
            if S3Manager._validate_json_object(s3_data):
                S3Manager.upload_json_object(
                    s3_data, manual_overrides_backup, s3_bucket
                )
            else:
                s3_data = _restore_backup()
        except Exception:
            _logger.error(f"Something went wrong reading {manual_overrides}")
            s3_data = _restore_backup()
        return s3_data

    @staticmethod
    def get_validated_xdre_replays(
        manual_overrides: str, manual_overrides_backup: str, s3_bucket: str
    ) -> Optional[dict]:
        def _restore_backup() -> Optional[dict]:
            s3_backup_data = S3Manager.get_json_object(
                manual_overrides_backup, s3_bucket
            )
            if not S3Manager._validate_json_object_xdre(s3_backup_data):
                _logger.error(
                    f"Neither {manual_overrides} nor {manual_overrides_backup} are valid."
                )
                return None

            _logger.warning(
                f"Restoring override file {manual_overrides} from {manual_overrides_backup}"
            )
            S3Manager.upload_json_object(s3_backup_data, manual_overrides, s3_bucket)
            return s3_backup_data

        # noinspection PyBroadException
        try:
            s3_data = S3Manager.get_json_object(manual_overrides, s3_bucket)
            if S3Manager._validate_json_object_xdre(s3_data):
                S3Manager.upload_json_object(
                    s3_data, manual_overrides_backup, s3_bucket
                )
            else:
                s3_data = _restore_backup()
        except Exception:
            _logger.error(f"Something went wrong reading {manual_overrides}")
            s3_data = _restore_backup()
        return s3_data

    @staticmethod
    def get_json_object(file_path, s3_bucket):
        try:
            s3_obj = S3Manager.s3_client.get_object(Bucket=s3_bucket, Key=file_path)
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "NoSuchKey":
                s3_data = None
            else:
                raise Exception(
                    f"Failed to Download Object from S3 due to Error {error.response['Error']['Code']}"
                ) from error

        else:
            s3_data = s3_obj["Body"].read().decode("utf-8")
        if not s3_data:
            return None

        return json.loads(s3_data)

    @staticmethod
    def upload_json_object(file_obj, file_path, s3_bucket):
        try:
            S3Manager.s3_client.put_object(
                Body=json.dumps(file_obj, indent=2), Bucket=s3_bucket, Key=file_path
            )
        except botocore.exceptions.ClientError as error:
            raise Exception(
                f"Failed to Upload Object to S3 due to Error {error.response['Error']['Code']}"
            ) from error

    @staticmethod
    def _validate_json_object(data: dict) -> bool:
        valid = True
        if data is None:
            return False
        for override_s, override_i in data.items():
            if not isinstance(override_s, str):
                _logger.error(
                    f"Invalid value for {override_s}: {override_i}, {override_s} must be a string not {type(override_s)}."
                )
                valid = False
            if not isinstance(override_i, int):
                _logger.error(
                    f"Invalid value for {override_s}: {override_i}, {override_i} must be a integer not {type(override_i)}."
                )
                valid = False
        return valid

    @staticmethod
    def _validate_json_object_xdre(data: dict) -> bool:
        valid = True
        if data is None:
            return False
        for override_s, override_obj in data.items():
            if not isinstance(override_s, str):
                _logger.error(
                    f"Invalid value for {override_s}: {override_obj}, {override_s} must be a string not {type(override_s)}."
                )
                valid = False
            if not isinstance(override_obj, object):
                _logger.error(
                    f"Invalid value for {override_s}: {override_obj}, {override_obj} must be a object not {type(override_obj)}."
                )
                valid = False
        return valid
